from hashlib import sha256

from hamt_sharding.consumable_hash import wrap_hash

def hash_fn(b: bytes) -> bytes:
    return sha256(b).digest()

def test_consumable_hash():
    val = 'some value'.encode()

    hash = wrap_hash(hash_fn)
    assert hash(val).take(0) == 0

    hash = wrap_hash(hash_fn)
    assert hash(val).take(10) == 110

    hash = wrap_hash(hash_fn)
    h = hash(val)
    for _ in range(100):
        result = h.take(10)
        assert result < 1024
        assert result > 0

    hash = wrap_hash(hash_fn)
    h = hash(val)
    h.take(10 * 100)
    h.untake(10 * 100)

    values = []
    hash = wrap_hash(hash_fn)
    h = hash(val)
    for _ in range(100):
        values.append(h.take(10))
    h.untake(10 * 100)
    for _ in range(100):
        result = h.take(10)
        assert result == values.pop(0)



from hashlib import sha256

from hamt_sharding.buckets import HAMTBucket

def hash_fn(value: bytes):
    return sha256(value).digest()

def test_hamt():
    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    assert bucket.get('unknown') is None
    
    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    bucket['key'] = 'value'
    assert bucket.get('key') == 'value'

    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    bucket['key'] = 'value'
    bucket['key'] = 'other value'
    assert bucket.get('key') == 'other value'

    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    del bucket['doesnt exist']

    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    bucket['key'] = 'value'
    del bucket['key']
    assert bucket.get('key') is None

    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    assert bucket.leaf_count() == 0
    for i in range(400):
        bucket[str(i)] = str(i)
    assert bucket.leaf_count() == 400

    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    for i in range(400):
        bucket[str(i)] = str(i)
    assert bucket.children_count() == 256

    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    assert bucket.only_child() is None

    bucket = HAMTBucket[str, str].create_hamt(hash_fn)
    for i in range(400):
        bucket[str(i)] = str(i)
    count = sum(1 for _ in bucket.each_leaf_series())
    assert count == 400

    def small_hash_fn(b) -> bytes:
        return hash_fn(b)[:2]

    bucket = HAMTBucket[str, str].create_hamt(small_hash_fn)
    for i in range(400):
        bucket[str(i)] = str(i)
    assert bucket['100'] == '100'

    bucket = HAMTBucket[bytes, int].create_hamt(small_hash_fn)
    bucket[b'test'] = 100
    assert bucket[b'test'] == 100
    assert list(bucket.keys()) == [b'test']
    assert list(bucket.values()) == [100]
    


import attr
from typing import TypeVar, Generic, Optional, Union, Generator, MutableMapping

from sparse_array import SparseArray
from .consumable_hash import InfiniteHash, InfiniteWrapper, HashFunction, wrap_hash

K = TypeVar('K', str, bytes)
V = TypeVar('V')

@attr.define(slots=True)
class HAMTBucketChild(Generic[K, V]):
    key: K
    value: V
    hash: InfiniteHash

@attr.define(slots=True)
class HAMTBucketPosition(Generic[K, V]):
    bucket: 'HAMTBucket[K, V]'
    pos: int
    hash: InfiniteHash
    existing_child: Optional[HAMTBucketChild[K, V]]

class HAMTBucket(MutableMapping[K, V]):
    def __init__(self, bits: int, hash: InfiniteWrapper, parent: Optional['HAMTBucket[K, V]'] = None, pos_at_parent: int = 0):
        self._bits = bits
        self._hash = hash
        self._pop_count = 0
        self._parent = parent
        self._pos_at_parent = pos_at_parent
        self._children = SparseArray[Union[HAMTBucket[K, V], HAMTBucketChild[K, V]]]()
        self.key: Optional[K] = None

    @classmethod
    def create_hamt(cls, hash_function: HashFunction, bits: int = 8):
        return cls(bits, wrap_hash(hash_function))

    def __setitem__(self, key: K, item: V):
        if not (isinstance(key, str) or isinstance(key, bytes)):
            raise TypeError(f'Bucket keys must be strings or bytes, not {type(key)}')
        
        place = self._find_new_bucket_and_pos(key)
        place.bucket._put_at(place, key, item)
        
    def __getitem__(self, key: K):
        if not (isinstance(key, str) or isinstance(key, bytes)):
            raise TypeError(f'Bucket keys must be strings or bytes, not {type(key)}')
        
        child = self._find_child(key)
        if child is not None:
            return child.value
        raise IndexError()
    
    def __len__(self):
        return self.leaf_count()
    
    def __iter__(self):
        for _, child in self._children.items():
            if isinstance(child, HAMTBucket):
                yield from child
            else:
                yield child.key
        
    def get(self, key: K, default: Optional[V] = None):
        try:
            return self[key]
        except IndexError:
            return default

    def __delitem__(self, key: str):
        if not (isinstance(key, str) or isinstance(key, bytes)):
            raise TypeError(f'Bucket keys must be strings or bytes, not {type(key)}')
        
        place = self._find_place(key)
        child = place.bucket._at(place.pos)
        if child is not None and child.key == key:
            place.bucket._del_at(place.pos)

    def leaf_count(self) -> int:
        result = 0
        for _, child in self._children.items():
            if isinstance(child, HAMTBucket):
                result += child.leaf_count()
            else:
                result += 1
        return result

    def children_count(self):
        return len(self._children)

    def only_child(self) -> Optional[HAMTBucketChild[K, V]]:
        return self._children.get(0)

    def each_leaf_series(self) -> Generator[HAMTBucketChild[K, V], None, None]:
        for _, child in self._children.items():
            if isinstance(child, HAMTBucket):
                yield from child.each_leaf_series()
            else:
                yield child

    @property
    def table_size(self):
        return pow(2, self._bits)

    def _find_child(self, key: K) -> Optional[HAMTBucketChild[K, V]]:
        result = self._find_place(key)
        child = result.bucket._at(result.pos)
        assert not isinstance(child, HAMTBucket)
        if child is not None and child.key == key:
            return child
        return None

    def _find_place(self, key: Union[K, InfiniteHash]) -> HAMTBucketPosition[K, V]:
        hash_value = self._hash(key.encode() if isinstance(key, str) else key)
        index = hash_value.take(self._bits)

        child = self._children.get(index)

        if isinstance(child, HAMTBucket):
            return child._find_place(hash_value)
        
        return HAMTBucketPosition(self, index, hash_value, child)
    
    def _find_new_bucket_and_pos(self, key: Union[str, InfiniteHash]) -> HAMTBucketPosition[K, V]:
        place = self._find_place(key)
        if place.existing_child is not None and place.existing_child.key != key:
            bucket = HAMTBucket(self._bits, self._hash, place.bucket, place.pos)
            place.bucket._put_object_at(place.pos, bucket)

            new_place = bucket._find_place(place.existing_child.hash)
            new_place.bucket._put_at(new_place, place.existing_child.key, place.existing_child.value)

            return bucket._find_new_bucket_and_pos(place.hash)
        return place

    def _put_at(self, place: HAMTBucketPosition[K, V], key: K, value: V):
        self._put_object_at(place.pos, HAMTBucketChild(key, value, place.hash))

    def _put_object_at(self, pos: int, object: Union['HAMTBucket[K, V]', HAMTBucketChild[K, V]]):
        if self._children.get(pos) is None:
            self._pop_count += 1
        self._children[pos] = object

    def _del_at(self, pos: int):
        assert pos >= 0
        if self._children.get(pos) is not None:
            self._pop_count -= 1
        del self._children[pos]
        self._level()

    def _at(self, pos: int):
        return self._children.get(pos)

    def _level(self):
        if self._parent is not None and self._pop_count <= 1:
            if self._pop_count == 1:
                only_child = next(self._children.values())
                if only_child is not None and not isinstance(only_child, HAMTBucket):
                    only_child.hash.untake(self._bits)
                    place = HAMTBucketPosition(self._parent, 
                                           self._pos_at_parent,
                                           only_child.hash,
                                           None)
                    self._parent._put_at(place, only_child.key, only_child.value)
            else:
                self._parent._del_at(self._pos_at_parent)



from hamt_sharding.consumable_buffer import ConsumableBuffer

def test_consumable():
    empty_buf = ConsumableBuffer(b'')
    assert empty_buf.take(0) == 0
    assert empty_buf.take(100) == 0
    assert empty_buf.take(1000) == 0

    buf = ConsumableBuffer(b'\0')
    assert buf.take(0) == 0
    assert buf.take(100) == 0
    assert buf.take(1000) == 0

    buf = ConsumableBuffer(bytes([0b11111111]))
    assert buf.take(0) == 0
    for _ in range(8):
        assert buf.take(1) == 1
    assert buf.take(1) == 0
    assert buf.take(100) == 0
    assert buf.take(1000) == 0

    buf = ConsumableBuffer(bytes([0xff, 0xff, 0xff]))
    assert buf.take(0) == 0
    for _ in range(24):
        assert buf.take(1) == 1
    assert buf.take(1) == 0

    buf = ConsumableBuffer(bytes([0xff, 0xff, 0xff]))
    for _ in range(12):
        assert buf.take(2) == 3
    assert buf.take(1) == 0

    buf = ConsumableBuffer(bytes([0xff, 0xff, 0xff]))
    assert buf.take(24) == 0b111111111111111111111111
    assert buf.take(1) == 0

    buf.untake(2)
    assert buf.take(2) == 3
    assert buf.take(1) == 0



from typing import Callable, Union

from .consumable_buffer import ConsumableBuffer

HashFunction = Callable[[bytes], bytes]
InfiniteWrapper = Callable[[Union['InfiniteHash', bytes]], 'InfiniteHash']

def wrap_hash(hash_function: HashFunction) -> InfiniteWrapper:
    def hashing(value: Union[bytes, InfiniteHash]):
        if isinstance(value, InfiniteHash):
            return value
        return InfiniteHash(value, hash_function)
    return hashing

class InfiniteHash:
    def __init__(self, value: bytes, hash_function: HashFunction):
        assert isinstance(value, bytes)
        self._value = value
        self._hash_function = hash_function
        self._depth = -1
        self._available_bits = 0
        self._current_buffer_index = 0
        self._buffers = list[ConsumableBuffer]()

    def take(self, bits: int) -> int:
        while self._available_bits < bits:
            self._produce_more_bits()
        result = 0
        while bits > 0:
            hash = self._buffers[self._current_buffer_index]
            avaliable = min(hash.available_bits(), bits)
            took = hash.take(avaliable)
            result = (result << avaliable) + took
            bits -= avaliable
            self._available_bits -= avaliable

            if hash.available_bits() == 0:
                self._current_buffer_index += 1
        return result
    
    def untake(self, bits: int):
        while bits > 0:
            hash = self._buffers[self._current_buffer_index]
            avaliable_for_untake = min(hash.total_bits() - hash.available_bits(), bits)
            hash.untake(avaliable_for_untake)
            bits -= avaliable_for_untake
            self._available_bits += avaliable_for_untake

            if self._current_buffer_index > 0 and hash.total_bits() == hash.available_bits():
                self._depth -= 1
                self._current_buffer_index -= 1

    def _produce_more_bits(self):
        self._depth += 1

        # JS Uint8Array.from([]) modulos with max value; implement that here
        value = self._value + bytes([self._depth % 0x100]) if self._depth > 0 else self._value
        hash_value = self._hash_function(value)
        buffer = ConsumableBuffer(hash_value)
        self._buffers.append(buffer)
        self._available_bits += buffer.available_bits()



START_MASKS = [
    0b11111111,
    0b11111110,
    0b11111100,
    0b11111000,
    0b11110000,
    0b11100000,
    0b11000000,
    0b10000000
]

STOP_MASKS = [
    0b00000001,
    0b00000011,
    0b00000111,
    0b00001111,
    0b00011111,
    0b00111111,
    0b01111111,
    0b11111111
]

def mask_for(start: int, length: int):
    return START_MASKS[start] & STOP_MASKS[min(length + start - 1, 7)]

def byte_bits_to_int(value: int, start: int, length: int):
    mask = mask_for(start, length)
    return (value & mask) >> start

class ConsumableBuffer:
    def __init__(self, value: bytes):
        self._value = value
        self._current_byte_pos = len(value) - 1
        self._current_bit_pos = 7

    def available_bits(self):
        return self._current_bit_pos + 1 + self._current_byte_pos * 8
    
    def total_bits(self):
        return len(self._value) * 8
    
    def _have_bits(self):
        return self._current_byte_pos >= 0
    
    def take(self, bits: int) -> int:
        result = 0
        while bits > 0 and self._have_bits():
            bite = self._value[self._current_byte_pos]
            avaliable_bits = self._current_bit_pos + 1
            taking = min(avaliable_bits, bits)
            value = byte_bits_to_int(bite, avaliable_bits - taking, taking)
            result = (result << taking) + value

            bits -= taking
            self._current_bit_pos -= taking

            if self._current_bit_pos < 0:
                self._current_bit_pos = 7
                self._current_byte_pos -= 1
        return result
    
    def untake(self, bits: int):
          self._current_bit_pos += bits
          while self._current_bit_pos > 7:
              self._current_bit_pos -= 8
              self._current_byte_pos += 1